#include <string.h>
#include <stdlib.h>
#include <signal.h>
#include <sys/types.h>
#include <unistd.h>
#include "parse.h"
#include "eval.h"

/* #define debug_eval */

static ef_elem * ef_top = NULL;

static fun * fun_top = NULL;

static arg * arg_top = NULL;

static node * eval_depend_on_type(node * n);

void print_argument_stack(void) {
  arg * tmp;
  printf("argument stack:\n");
  for (tmp=arg_top;tmp!=NULL;tmp=tmp->next) {
    printf("'%s: %s\n",tmp->symbol,type_lookup(tmp->n->type));
  }
  printf("end of argument stack\n");
}

static void register_argument(char * symbol, node * n) {
  arg * a = xmalloc(sizeof(arg));
  a->symbol = symbol;
  a->n = n;
  a->next = arg_top;
  arg_top = a;
#ifdef debug_eval
  print_argument_stack();
#endif
  /*
  if (n->type == SYMBOL) {
    kill(getpid(),SIGSEGV);
  }
  */
#ifdef debug_eval
  printf("argument '%s registered (%d)\n",symbol,n->int_data);
  print_argument_stack();
#endif
}

node * get_argument(char * symbol) {
  arg * tmp;
  for (tmp=arg_top;tmp!=NULL;tmp=tmp->next) {
    if (strcasecmp(symbol,tmp->symbol)==0) {
      return tmp->n;
    }
  }
  return NULL;
}

static void unregister_argument(char * symbol) {
  arg * tmp = arg_top;
  if (strcasecmp(tmp->symbol,symbol)==0) {
    arg_top = tmp->next;
#ifdef NO_BOEHM_GC
    /* free(tmp); */
#endif
    return;
  }
  for (;tmp->next!=NULL;tmp=tmp->next) {
    if (strcasecmp(tmp->next->symbol,symbol)==0) {
      /* arg * tmp2 = tmp->next; */
      tmp->next = tmp->next->next;
#ifdef NO_BOEHM_GC
      /* free(tmp2); */
#endif
      return;
    }
  }
  /* nothing found, nothing removed -- should we spit out a warning? */
}

node * eval_subtree(node * n) {
  if (n) {
    if (n->symbol_data) {
      node * (*eval_func)(char *, node **, int) = (node * (*)(char *, node **, int))lookup_eval_func(n->symbol_data);
      if (eval_func==NULL) {
        printf("unknown function %s\n",n->symbol_data);
        /* raise(SIGSEGV); */
        exit(1);
#ifdef debug_eval
        printf("eval_subtree: eval_func is NULL\n");
#endif
      }
      return eval_func(n->symbol_data,n->childs,n->child_count);
    }
    return n; /* without a symbol, we return the tree itself */
  }
  return NULL;
}

static node * eval_int(node * n) {
  if (n && n->type==INT) {
    return n;
  }
  /* maybe put automatic conversion here? */
  printf("eval_int: expected INT, got %s\n",type_lookup(n->type));
  exit(1);
  return NULL;
}

static node * eval_string(node * n) {
  if (n && n->type==STRING) {
    return n;
  }
  /* dito: put automatic conversion here? */
  printf("eval_string: expected STRING, got %s\n",type_lookup(n->type));
  exit(1);
  return NULL;
}

static node * eval_symbol(node * n) {
  if (n && n->type==SYMBOL) {
    arg * x;
    for (x=arg_top;x!=NULL;x=x->next) {
      if (strcasecmp(n->symbol_data,x->symbol)==0) {
#ifdef debug_eval
        printf("eval_symbol: %s: type of x->n is %s\n",n->symbol_data,type_lookup(x->n->type));
#endif
        return eval_depend_on_type(x->n);
      }
    }
#ifdef debug_eval
    printf("eval_symbol: return symbol itself\n");
#endif
    return n; /* when no symbol can be found in the list, return symbol itself */
  }
  return  NULL;
}

static node * eval_depend_on_type(node * n) {
  if (n) {
#ifdef debug_eval
    printf("eval_depend_on_type: type is %s\n",type_lookup(n->type));
#endif
    switch (n->type) {
    case SUB_TREE:
      return eval_subtree(n);
    case INT:
      return eval_int(n);
    case STRING:
      return eval_string(n);
    case SYMBOL:
      return eval_symbol(n);
    default:
      return NULL;
    }
  }
  return NULL;
}

void * lookup_eval_func(char * fn) {
  ef_elem * x;
#ifdef debug_eval
  printf("lookup_eval_func: trying to match %s\n",fn);
#endif
  if (!fn) {
    return NULL;
  }
  for (x=ef_top;x!=NULL;x=x->next) {
    if (strcasecmp(x->id,fn)==0) {
      return x->f;
    }
  }
  return NULL;
}

static node * eval_func_plus(char * symbol, node ** childs, int count) {
  int result = 0, i;
  node * n;
  node * x = new_node(0);
  for (i=0;i<count;i++) {
    node * o;
    n = childs[i];
    o = eval_depend_on_type(n);
    if (!o) {
      printf("plus: evaluated subtree is NULL\n");
      exit(1);
    }
    if (o->type == INT) {
      result += o->int_data;
    } else {
      printf("plus: expected INT, got %s\n",type_lookup(o->type));
      exit(1);
    }
  }
  x->type = INT;
  x->int_data = result;
  return x;
}

static node * eval_func_mul(char * symbol, node ** childs, int count) {
  int result = 1, i;
  node * n;
  node * x = new_node(0);
  for (i=0;i<count;i++) {
    node * o;
    n = childs[i];
    o = eval_depend_on_type(n);
    if (!o) {
      printf("mul: evaluated subtree is NULL\n");
      exit(1);
    }
    if (o->type == INT) {
      result *= o->int_data;
    } else {
      printf("mul: expected INT, got %s\n",type_lookup(o->type));
      exit(1);
    }
  }
  x->type = INT;
  x->int_data = result;
  return x;
}

static node * eval_func_minus(char * symbol, node ** childs, int count) {
  node * n1, * n2;
  node * result;
  if (count!=2) {
    printf("minus: expected 2 arguments, got %d\n",count);
    exit(1);
  }
#ifdef debug_eval
  printf("argument 1: type is %s\n",type_lookup(childs[0]->type));
  printf("argument 2: type is %s\n",type_lookup(childs[1]->type));
#endif
  n1 = eval_depend_on_type(childs[0]);
  n2 = eval_depend_on_type(childs[1]);
  if (n1->type!=INT) {
    printf("minus: argument 1 is %s, expected INT\n",type_lookup(n1->type));
    exit(1);
  }
  if (n2->type!=INT) {
    printf("minus: argument 2 is %s, expected INT\n",type_lookup(n2->type));
    exit(1);
  }
  result = new_node(0);
  result->type = INT;
  result->int_data = n1->int_data - n2->int_data;
  return result;
}


static node * eval_func_div(char * symbol, node ** childs, int count) {
  node * n1, * n2;
  node * result;
  if (count!=2) {
    printf("div: expected 2 arguments, got %d\n",count);
    exit(1);
  }
  n1 = eval_depend_on_type(childs[0]);
  n2 = eval_depend_on_type(childs[1]);
  if (n1->type!=INT) {
    printf("div: argument 1 is %s, expected INT\n",type_lookup(n1->type));
    exit(1);
  }
  if (n2->type!=INT) {
    printf("div: argument 2 is %s, expected INT\n",type_lookup(n2->type));
    exit(1);
  }
  result = new_node(0);
  result->type = INT;
  result->int_data = n1->int_data / n2->int_data;
  return result;
}

static node * eval_func_userdefined(char * symbol, node ** childs, int count) {
  fun * tmp;
  node * fhead;
  node ** fbody;
  node * result = NULL;
  int bodycount;
  int i;
  for (tmp=fun_top;tmp!=NULL;tmp=tmp->next) {
    if (strcasecmp(symbol,tmp->name)==0) {
      break;
    }
  }
  if (!tmp) {
    printf("unable to find a function definition for %s\n",symbol);
    exit(1);
  }
  fhead = tmp->head;
  if (count!=fhead->child_count) {
    printf("argument count mismatch: expected %d arguments for %s, got %d arguments\n",fhead->child_count,symbol,count);
    exit(1);
  }
  for (i=0;i<fhead->child_count;i++) {
    node * x = NULL;
    if (fhead->childs[i]->type != SYMBOL) {
      printf("argument declaration %d of %s is not a symbol\n",i,symbol);
      exit(1);
    }
    if (childs[i]->type==SYMBOL) {
      x = get_argument(childs[i]->symbol_data);
      if (!x) {
        x = childs[i]; /* XXX */
      }
    } else {
      x = childs[i];
    }
    if (x->type==SUB_TREE) {
      x = eval_subtree(x);
    }
    register_argument(fhead->childs[i]->symbol_data,x); /* XXX */
  }
  fbody = tmp->body;
  bodycount = tmp->bodycount;
  for (i=0;i<bodycount;i++) {
    result = eval_depend_on_type(fbody[i]); /* memory leak!!! */
  }
  for (i=0;i<fhead->child_count;i++) {
    unregister_argument(fhead->childs[i]->symbol_data);
  }
  return result;
}


static void register_function(char * fname, node * fhead, node ** fbody, int bodycount) {
  fun * function = xmalloc(sizeof(fun));
  function->name = strdup(fname);
  function->head = fhead;
  function->body = fbody;
  function->bodycount = bodycount;
  function->next = fun_top;
  fun_top = function;
  /* register user-defined function;
   * this is handled by eval_func_userdefined
   */
  add_eval_func(function->name,(void *)eval_func_userdefined);
}

static node * eval_func_defun(char * symbol, node ** childs, int count) {
  node * fhead;
  if (count<2) {
    printf("defun: function body missing\n");
    exit(1);
  }
  fhead = childs[0];
  if (fhead->symbol_data) {
    register_function(fhead->symbol_data,fhead,childs+1,count-1);
  } else {
    printf("defun: empty function head\n");
    exit(1);
  }
  return NULL;
}

static node * eval_func_if(char * symbol, node ** childs, int count) {
  node * cond = NULL, * ifyes = NULL, * ifno = NULL;
  if (count<2) {
    printf("if: too few arguments\n");
    exit(1);
  }
  cond = childs[0];
  ifyes = childs[1];
  if (count==3) {
    ifno = childs[2];
  }
  cond = eval_depend_on_type(cond);
  if (!cond) {
    printf("if: condition evaluated to NULL\n");
    exit(1);
  }
  if (cond->type!=INT) {
    printf("if: condition must be of type INT\n");
    exit(1);
  }
  if (cond->int_data) {
    return eval_depend_on_type(ifyes);
  } else if (ifno) {
    return eval_depend_on_type(ifno);
  }
  return NULL; /* shouldn't be */
}

node * eval_func_not(char * symbol, node ** childs, int count) {
  node * n;
  if (count!=1) {
    printf("not: need one argument\n");
    exit(1);
  }
  n = eval_depend_on_type(childs[0]);
  if (n->type==INT) {
    node * x = new_node(0);
    x->type = INT;
    x->int_data = n->int_data?0:1; /* or !n->int_data ? */
    return x;
  }
  printf("not: need type INT as argument 0, got %s\n",type_lookup(n->type));
  exit(1);
  return NULL;
}

static void do_print_subtree(node * x) {
  if (x) {
    switch (x->type) {
    case SUB_TREE: {
      int i;
      printf("(%s ",x->symbol_data?x->symbol_data:"");
      for (i=0;i<x->child_count;i++) {
        do_print_subtree(x->childs[i]);
      }
      printf(")");
      break;
    }
    case INT:
      printf("%d ",x->int_data);
      break;
    case STRING:
      printf("\"%s\" ",x->str_data);
      break;
    case SYMBOL:
      printf("'%s ",x->symbol_data);
      break;
    default:
      printf("do_print_subtree: unknown type %d\n",x->type);
      exit(1);
    }
  }
}

node * eval_func_print(char * symbol, node ** childs, int count) {
  int i;
  for (i=0;i<count;i++) {
    node * x = eval_depend_on_type(childs[i]);
    if (x) {
      switch (x->type) {
      case SUB_TREE:
        printf("<subtree:0x%p",x);
        do_print_subtree(x);
        printf(">");
        break;
      case STRING:
        printf("%s",x->str_data);
        break;
      case SYMBOL:
        printf("'%s",x->symbol_data);
        break;
      case INT:
        printf("%d",x->int_data);
        break;
      default:
        printf("print: unknown type %d\n",x->type);
        exit(1);
      }
    }
  }
  return NULL;
}

node * eval_func_nl(char * symbol, node ** childs, int count) {
  static node n = { NULL, 0, STRING, 0, "\n", NULL };
  return &n;
}

node * eval_func_car(char * symbol, node ** childs, int count) {
  if (count>0) {
    node * x = eval_depend_on_type(childs[0]);
    if (x && x->type==SUB_TREE) {
      if (x->symbol_data) {
        x = eval_subtree(x);
      }
      if (x->childs) return x->childs[0];
    }
  }
  return NULL;
}

node * eval_func_cdr(char * symbol, node ** childs, int count) {
  node * retnode;
  node * node2cdr;
  int i;
  if (count==0) {
    return NULL;
  }
  node2cdr = eval_depend_on_type(childs[0]);
  if (node2cdr->child_count==0) {
    return NULL;
  }
  retnode = new_node(node2cdr->child_count-1);
  retnode->type = SUB_TREE;
  retnode->childs = xmalloc((node2cdr->child_count-1)*sizeof(node));
  for (i=1;i<node2cdr->child_count;i++) {
    retnode->childs[i-1] = node2cdr->childs[i];
  }
  return retnode;
}

node * eval_func_gt(char * symbol, node ** childs, int count) {
  node * n1, * n2, * retnode;
  int retval = 0;
  if (count!=2) {
    printf("gt: need two arguments\n");
    exit(1);
  }
  n1 = eval_depend_on_type(childs[0]);
  n2 = eval_depend_on_type(childs[1]);
  if (n1->type != INT || n2->type!=INT) {
    printf("gt: type mismatch: both arguments need to be INT\n");
    exit(1);
  }
  if (n1->int_data > n2->int_data) {
    retval = 1;
  }
  retnode = new_node(0);
  retnode->type = INT;
  retnode->int_data = retval;
  return retnode;
}

node * eval_func_lt(char * symbol, node ** childs, int count) {
  node * n1, * n2, * retnode;
  int retval = 0;
  if (count!=2) {
    printf("lt: need two arguments\n");
    exit(1);
  }
  n1 = eval_depend_on_type(childs[0]);
  n2 = eval_depend_on_type(childs[1]);
  if (n1->type != INT || n2->type!=INT) {
    printf("lt: type mismatch: both arguments need to be INT\n");
    exit(1);
  }
  if (n1->int_data < n2->int_data) {
    retval = 1;
  }
  retnode = new_node(0);
  retnode->type = INT;
  retnode->int_data = retval;
  return retnode;
}

node * eval_func_and(char * symbol, node ** childs, int count) {
  int retval;
  node * x;
  int i;
  if (count<2) {
    printf("and: need at least two argument\n");
    exit(1);
  }
  x = eval_depend_on_type(childs[0]);
  if (x->type!=INT) {
    printf("and: type mismatch: argument 0 is not INT\n");
    exit(1);
  }
  retval = x->int_data;
  for (i=1;i<count;i++) {
    x = eval_depend_on_type(childs[i]);
    if (x->type!=INT) {
      printf("and: type mismatch: argument %d is not INT\n",i);
      exit(1);
    }
    retval = retval && x->int_data;
  }
  x = new_node(0);
  x->type = INT;
  x->int_data = retval;
  return x;
}

node * eval_func_or(char * symbol, node ** childs, int count) {
  int retval;
  node * x;
  int i;
  if (count<2) {
    printf("or: need at least two argument\n");
    exit(1);
  }
  x = eval_depend_on_type(childs[0]);
  if (x->type!=INT) {
    printf("or: type mismatch: argument 0 is not INT\n");
    exit(1);
  }
  retval = x->int_data;
  for (i=1;i<count;i++) {
    x = eval_depend_on_type(childs[i]);
    if (x->type!=INT) {
      printf("or: type mismatch: argument %d is not INT\n",i);
      exit(1);
    }
    retval = retval || x->int_data;
  }
  x = new_node(0);
  x->type = INT;
  x->int_data = retval;
  return x;
}

node * eval_func_eq(char * symbol, node ** childs, int count) {
  node * n1, * n2, * retnode;
  int retval = 0;
  if (count!=2) {
    printf("eq: need two arguments\n");
    exit(1);
  }
  n1 = eval_depend_on_type(childs[0]);
  n2 = eval_depend_on_type(childs[1]);
  if (n1->type != n2->type) {
    printf("eq: type mismatch. can't compare %s and %s\n",type_lookup(n1->type),type_lookup(n2->type));
    exit(1);
  }
  switch (n1->type) {
  case SUB_TREE:
    /* unimplemented - how should it be done? */
    retval = 0; break;
  case INT:
    if (n1->int_data == n2->int_data) {
      retval = 1;
    } else {
      retval = 0;
    }
    break;
  case STRING:
    if (strcmp(n1->str_data,n2->str_data)==0) {
      retval = 1;
    } else {
      retval = 0;
    }
    break;
  case SYMBOL:
    if (strcasecmp(n1->symbol_data,n2->symbol_data)==0) {
      retval = 1;
    } else {
      retval = 0;
    }
    break;
  default:
    printf("unknown type %d\n",n1->type);
  }
  retnode = new_node(0);
  retnode->int_data = retval;
  retnode->type = INT;
#ifdef debug_eval
  printf("eq: return value is %d\n",retval);
#endif
  return retnode;
}

node * eval_func_isnull(char * symbol, node ** childs, int count) {
  node * x, * y;
  if (count!=1) {
    printf("isnull: only one argument, please\n");
    exit(1);
  }
  if (childs[0] && childs[0]->type==SUB_TREE && childs[0]->symbol_data) {
    y = eval_subtree(childs[0]);
  } else {
    y = childs[0];
  }
  x = new_node(0);
  x->type = INT;
  x->int_data = y?0:1;
  return x;
}

node * eval_func_eval(char * symbol, node ** childs, int count) {
  int i;
  node * retval = NULL;
  for (i=0;i<count;i++) {
    retval = eval_depend_on_type(childs[i]);
  }
  return retval;
}

node * eval_func_load(char * symbol, node ** childs, int count) {
  int i;
  for (i=0;i<count;i++) {
    if (childs[i]) {
      node * x = eval_depend_on_type(childs[i]);
      if (x->type==STRING) {
        load_and_run(x->str_data,NULL);
      } else {
        printf("load: type mismatch: expected type STRING for argument %d, got %s\n",i,type_lookup(x->type));
        exit(1);
      }
    }
  }
  return NULL;
}

node * eval_func_exit(char * symbol, node ** childs, int count) {
  if (count==0) {
    exit(0);
    return NULL; /* NOTREACHED */
  }
  if (count==1) {
    node * x = eval_depend_on_type(childs[0]);
    if (x && x->type==INT) {
      exit(x->int_data);
    } else {
      printf("exit: need either no or exactly one argument of type INT\n");
      exit(1);
    }
  }
  printf("exit: expected no or one argument, got %d\n",count);
  exit(1);
}

node * eval_func_cons(char * symbol, node ** childs, int count) {
  node * el2ins, * list2cons, * n;
  int i;
  if (count!=2) {
    printf("cons: wrong argument count\n");
    exit(1);
  }
  el2ins = childs[0];
  list2cons = eval_depend_on_type(childs[1]);
  if (list2cons->type!=SUB_TREE) {
    printf("cons: argument 2 must be of type SUB_TREE\n");
    exit(1);
  }
  n = new_node(list2cons->child_count+1);
  n->type = list2cons->type;
  n->child_count = list2cons->child_count+1;
  n->int_data = list2cons->int_data;
  n->str_data = list2cons->str_data?strdup(list2cons->str_data):NULL;
  n->symbol_data = list2cons->symbol_data?strdup(list2cons->symbol_data):NULL;
  for (i=0;i<list2cons->child_count;++i) {
    n->childs[i+1] = list2cons->childs[i];
  }
  n->childs[0] = el2ins;
  return n;
}

node * eval_func_at(char * symbol, node ** childs, int count) {
  node * l, * x;
  if (count!=2) {
    printf("at: wrong argument count\n");
    exit(1);
  }
  l = eval_depend_on_type(childs[0]);
  x = eval_depend_on_type(childs[1]);
  if (l->type!=SUB_TREE) {
    printf("at: argument 1 must be of type SUB_TREE\n");
    exit(1);
  }
  if (x->type!=INT) {
    printf("at: argument 2 must be of type INT\n");
    exit(1);
  }
  if (x->int_data >= 0 && x->int_data < l->child_count) {
    return l->childs[x->int_data];
  }
  printf("at: out of range\n");
  exit(1);
  return NULL;
}

node * eval_func_size(char * symbol, node ** childs, int count) {
  node * n;
  if (count!=1) {
    printf("size: wrong argument count\n");
    exit(1);
  }
  n = eval_depend_on_type(childs[0]);
  if (n && n->type==SUB_TREE) {
    node * x = new_node(0);
    x->type = INT;
    x->int_data = n->child_count;
    return x;
  } else {
    printf("size: argument 1 must be of type SUB_TREE\n");
    exit(1);
  }
  return NULL;
}

void init_eval_system(void) {
  /* better put this into an array and iterate over it */
  add_eval_func("plus",(void *)eval_func_plus);
  add_eval_func("minus",(void *)eval_func_minus);
  add_eval_func("mul",(void *)eval_func_mul);
  add_eval_func("div",(void *)eval_func_div);
  add_eval_func("defun",(void *)eval_func_defun);
  add_eval_func("if",(void *)eval_func_if);
  add_eval_func("eq",(void *)eval_func_eq);
  add_eval_func("not",(void *)eval_func_not);
  add_eval_func("print",(void *)eval_func_print);
  add_eval_func("nl",(void *)eval_func_nl);
  add_eval_func("car",(void *)eval_func_car);
  add_eval_func("cdr",(void *)eval_func_cdr);
  add_eval_func("isnull",(void *)eval_func_isnull);
  add_eval_func("eval",(void *)eval_func_eval);
  add_eval_func("gt",(void *)eval_func_gt);
  add_eval_func("lt",(void *)eval_func_lt);
  add_eval_func("or",(void *)eval_func_or);
  add_eval_func("and",(void *)eval_func_and);
  add_eval_func("load",(void *)eval_func_load);
  add_eval_func("exit",(void *)eval_func_exit);
  add_eval_func("cons",(void *)eval_func_cons);
  add_eval_func("at",(void *)eval_func_at);
  add_eval_func("size",(void *)eval_func_size);
}

void add_eval_func(char * id, void * fptr) {
  ef_elem * x = xmalloc(sizeof(ef_elem));
  x->f = fptr;
  x->id = strdup(id);
  x->next = ef_top;
  ef_top = x;
}

